// 2011.09.08: Header is included from util.h.
//#include <stdio.h>
//#include <math.h>
//#include <stdlib.h>
#include <string.h>
//#include "mkl_lapack.h"
//#include "mkl_cblas.h"

#include "util.h"


void printUsage(){
printf("\n");
printf("Usage: gmm [options] -n N -i INPUT -o OUTPUT [> logfile]\n");
printf("\n");
printf("Fit a Bayesian Gaussian mixture model to SNP array signal intensity data.\n");
printf("\n");
printf("Options:\n");
printf("  -h, --help        Print short help message and exit\n");
printf("  --version         Print version info and exit\n");
printf("  -i                INPUT file name of signal intensity data (binary, double-precision)\n");
printf("  -n                Sample size N\n");
printf("  -s                Subsamples specified by a series of 0/1 values (binary, integer)\n");
printf("  -o                OUTPUT file name of estimated parameters and QC statistics\n");
printf("\n");
printf("                    Value:                 Field(s):\n");
printf("                    Mu                     1-%d\n", K*2);
printf("                    Sigma                  %d-%d\n", K*2+1, K*6);
printf("                    ASCN Genotype Freq.    %d-%d\n", K*6+1, K*7);
printf("                    ASCN Haplotype Freq.   %d-%d\n", K*7+1, K*7+J);
printf("                    Convergence Status     %d \n", K*7+J+1);
printf("                      0 : Successfully converged      \n");
printf("                      1 : Cluster overlapping         \n");
printf("                      2 : Skewed lattice structure    \n");
printf("                      3 : 1 & 2                       \n");
printf("                      4 : Iteration limit exceeded    \n");
printf("                      99: No data                     \n");
//printf("                    Theta                  %d \n", K*7+J+2);
//printf("                    Lattice                %d \n", K*7+J+3);
//printf("                    N of cluster overlap   %d \n", K*7+J+4);
printf("                    Call Rate              %d \n", K*7+J+2);
printf("                    Log10 HWE P-value      %d \n", K*7+J+3);
printf("                    Estimated Error Rate   %d \n", K*7+J+4);
printf("                    Posterior Probability  %d \n", K*7+J+5);
printf("                    Mesh resolution        %d \n", K*7+J+6);
printf("                    -------------------------------\n");
printf("                    Total                  %d \n", K*7+J+6);
printf("\n");
printf("  -t, --text        OUTPUT as text\n");
printf("  -v, --verbose     Print more information about progress\n");
//printf("  -q, --quiet           Don't print any message\n");
printf("\n");
printf("Report bugs to <kumasaka@src.riken.jp>.\n\n");
}



int main(int argc, char **argv){
	int itr,i,j,k,flag;
	int inputfileid=0;
	int qcstatslen=6;
	FILE *fp, *op,  *fsubset;
	int verbose, text;
	verbose=0;
	text=0;
	
	if(argc<2){printUsage(); return 0;}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-h")==0 || strcmp(argv[i],"--help")==0){printUsage(); return 0;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"--version")==0){printVersion(); return 0;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-v")==0 || strcmp(argv[i],"--verbose")==0){verbose=1;}}
	//for(i=0; i<argc; i++){if(strcmp(argv[i],"-q")==0 || strcmp(argv[i],"--quite")==0){quite=1;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-t")==0 || strcmp(argv[i],"--text")==0){text=1;}}
	
	printVersion();
	printf("User command:\n");
	for(i=0; i<argc; i++){printf("%s ", argv[i]);}
	printf("\n\n");
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-i")==0 && argc>i+1){if((fp=fopen(argv[i+1],"rb"))==NULL){printf("No intensity file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("Intensity file:  %s\n", argv[i+1]);flag=1;inputfileid = i+1;};break;}}
	if(flag==0){printf("No signal intensity file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-o")==0 && argc>i+1){if((op=fopen(argv[i+1],"wb"))==NULL){printf("No output file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("Output file: %s\n", argv[i+1]);flag=1;};break;}}
	if(flag==0){printf("No output file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-s")==0 && argc>i+1){if((fsubset = fopen(argv[i+1],"rb"))==NULL){printf("No subset file exists!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("Subset file: %s\n", argv[i+1]);flag=1;};break;}}
	if(flag==0){printf("Subset file: no\n"); fsubset=NULL;}
	
	printf("\n\n");
	
	
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-n")==0 && argc>i+1){if(sscanf(argv[i+1], "%d", &N)==0){printf("Sample size is invalid!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("Sample size: %d\n", N);}; flag=1;}}
	if(flag==0){printf("Sample size is missing!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	if(N<15){printf("Sample size is too small!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	
	int nl=0;
	ldiv_t nlt = ncol(argv[inputfileid], 2*N);
	if(nlt.rem==0){
		nl = (int)nlt.quot;
		printf("Marker size: %d\n", nl);
	}else{
		printf("Sample size is inconsistent with the intensity file size! (size of %s = 8 * %ld * %d + %ld) \n\nPlatinumCNV terminated...\n\n",argv[inputfileid],nlt.quot,N*2,nlt.rem);exit(EXIT_FAILURE);
	}
	
	printf("\n\n");
	
	
	double lkhd;
	
	double *Theta;
	double *Theta15;
	double *eTheta15;
	double *Theta21;

	double *Mu;
	double *Sigma;
	double *Mutemp;
	double *Sigmatemp;
	double *xt;
	double *freq;

	double* Z6;
	double* Z15;
	double* Z21;
	double* d;
	double* x0; 
	double* x;
	double* w;
	int* wtmp;
	int* tab;
	int* subset;

	int N0, Nw;
	int flag1;
	
	double cr[2];
	int status=0;
	
	
	
	Theta = (double*)calloc(J, sizeof(double));
	Theta15 = (double*)calloc(K, sizeof(double));
	eTheta15 = (double*)calloc(K, sizeof(double));
	Theta21 = (double*)calloc(J*(J+1)/2, sizeof(double));
	
	Mu = (double*)calloc(K*2, sizeof(double));
	Sigma = (double*)calloc(4*K, sizeof(double));
	Mutemp = (double*)calloc(K*2, sizeof(double));
	Sigmatemp = (double*)calloc(K*4, sizeof(double));
	xt = (double*)calloc(2, sizeof(double));
	freq = (double*)calloc(K, sizeof(double));
	
	
	hess = (double*)calloc(K*4*K, sizeof(double));
	
	subset = (int*)calloc(N, sizeof(int));
	x0 = (double*)calloc(N*2, sizeof(double));
	x  = (double*)calloc(N*2,  sizeof(double));
	w = (double*)calloc(N,  sizeof(double));
	wtmp = (int*)calloc(N,  sizeof(int));
	tab = (int*)calloc(N,  sizeof(int));
	
	Z6 = (double*)calloc(N*J, sizeof(double));
	Z15 = (double*)calloc(N*K, sizeof(double));
	d = (double*)calloc(N, sizeof(double));
	Z21 = (double*)calloc(N*(J+1)*J/2, sizeof(double));
	X1 = (double*)calloc(N*2, sizeof(double));
	X2 = (double*)calloc(N*2, sizeof(double));
	X3 = (double*)calloc(N*2, sizeof(double));
	
	uvn = (double*)calloc(N, sizeof(double));
	lk = (double*)calloc(500, sizeof(double));
	uvk = (double*)calloc(K, sizeof(double));
	uvj = (double*)calloc(J*(J+1)/2, sizeof(double));
	
	for(i=0; i<N; i++){uvn[i] = 1.0;}
	for(i=0; i<K; i++){uvk[i] = 1.0;}
	for(i=0; i<J*(J+1)/2; i++){uvj[i] = 1.0;} 
	
	if(X1 == NULL || X2 == NULL || X3 == NULL || Z6 == NULL || Z15 == NULL ||Z21 == NULL || uvn == NULL || x == NULL || x0 == NULL || w == NULL || wtmp == NULL || tab == NULL || subset == NULL){printf("Memory allocation failure...\nSample size (%d) is too large!\n", N); exit(EXIT_FAILURE);}else{printf("Memory allocated successfully...\n\n");}

	
	
	
	if(fsubset==NULL){
		//printf("odata not found!\n");exit(EXIT_FAILURE);
		for(i=0; i<N; i++){subset[i] = 1;}
		N0 = N;
	}else{ 
		//printf("subset=T\n");
		N0 = N;
		fread(subset, sizeof(int), N0, fsubset);
		N = 0;
		for(i=0; i<N0; i++){if(subset[i]==1){N++;}}
	}
	printf("N of subjects used: %d/%d\n\n\n", N, N0);
	
	
	
	
	
//###############   Scaling  #####################
	//cblas_dscal(K*2, 0.8, Mu0, 1);
	//cblas_dscal(K*4, 0.64, Sigma0, 1);
	//cblas_dscal(K*2, 1.25, Mu0, 1);
	//cblas_dscal(K*4, 1.5625, Sigma0, 1);
//for(i=0; i<K; i++){Mu0[i*2] = 1.0129123*pow(Mu0[i*2],0.8675247);}
//################################################
//vprint(Mu0,K*2);
//vprint(Sigma0,K*4);
	alpha0 = cblas_dasum(J, alpha, 1);
	cblas_dcopy(K*2, Mu0, 1, Mutemp, 1);
	cblas_dcopy(K*4, Sigma0, 1, Sigmatemp, 1);
	
	
	

	for(itr=0; itr<nl; itr++){
		printf("Marker %d\t",itr+1);
		fread(x0, sizeof(double), 2*N0, fp);
		j=0;
// 2011.09.08: Initialize q1, q2 to 0.0.
		double q1 = 0.0, q2 = 0.0;
		double bound = 12.0;
// 2011.09.08: Initialize mesh_ratio, mesh_ratio0 to 0.0.
		double mesh_ratio = 100.0, mesh_ratio0 = 100.0;
		if(N>3000){
			mesh_ratio0 = mesh_ratio = 100.0;
			
			int ii;
			clear(x,N0*2);
			for(ii=0; ii<25; ii++){
				q1 = q2 = 0.0;
				clear(w,N0);
				clearInt(tab,N0);
				clearInt(wtmp,N0);
				Nw=0;
				j=0;
				//printf("m=%lf\n", mesh_ratio);
				for(k=0; k<N0; k++){
					if(subset[k]==1){
						//xt[0] = floor(1.006711*pow(x0[2*k],0.9317417)*mesh_ratio);
						//xt[1] = floor(pow(x0[2*k+1],1.073259)/1.006711*mesh_ratio);
						xt[0] = floor(x0[2*k]*mesh_ratio);
						xt[1] = floor(x0[2*k+1]*mesh_ratio);
						if(xt[0]>bound*mesh_ratio){xt[0]=bound*mesh_ratio;}
						if(xt[0]<0.0)  {xt[0]=0.0;}
						if(xt[1]>bound*mesh_ratio){xt[1]=bound*mesh_ratio;}
						if(xt[1]<0.0)  {xt[1]=0.0;}
						if(isnan(xt[0])==0 && isnan(xt[1])==0){
							q1 += xt[0];
							q2 += xt[1];
							wtmp[j++] = ((int)xt[0])+((int)xt[1])*((int)bound)*((int)mesh_ratio);
						}
					}
				}
				
				table(wtmp, j, tab, w, &Nw);
				//printf("j=%d Nw=%d\n",j, Nw);
				if(Nw>3000 && Nw<5000){
					for(i=0; i<Nw; i++){
						div_t dd;
						dd = div(tab[i], ((int)bound)*((int)mesh_ratio));
						x[i*2]   = ((double) dd.rem + 0.5)/mesh_ratio;
						x[i*2+1] = ((double) dd.quot + 0.5)/mesh_ratio;
					}
					q1 /= ((double)j)*mesh_ratio;
					q2 /= ((double)j)*mesh_ratio;
					break;
				}else{
					double tmp = mesh_ratio;
					if(Nw>(5000-1)){
						if(ii==0){mesh_ratio /= 1.6;}else{
							if(mesh_ratio0<mesh_ratio){// ---m0----|   |---m----
								mesh_ratio = (mesh_ratio+mesh_ratio0)/2.0;
								mesh_ratio0 = tmp;
							}else{// ---|   |---m----m0----
								mesh_ratio -= (mesh_ratio0-mesh_ratio);
								mesh_ratio0 = tmp;
								
							}
						}
					}else{
						if(ii==0){mesh_ratio *= 1.6;}else{
							if(mesh_ratio0<mesh_ratio){// ---m0---m---|   |----
								mesh_ratio += (mesh_ratio-mesh_ratio0);
								mesh_ratio0 = tmp;
							}else{// ---m----|   |---m0----
								mesh_ratio = (mesh_ratio+mesh_ratio0)/2.0;
								mesh_ratio0 = tmp;
							}
						}
					}
				}
			}
		}else{
			int npe=0.0;
			for(k=0; k<N0; k++){
				if(subset[k]==1){
					//xt[0] = 1.006711*pow(x0[2*k],0.9317417);
					//xt[1] = pow(x0[2*k+1],1.073259)/1.006711;
					xt[0] = x0[2*k];
					xt[1] = x0[2*k+1];
					
					if(isnan(xt[0])==0 && isnan(xt[1])==0){
						if(xt[0]>bound){xt[0]=bound;};
						if(xt[0]<0.0)  {xt[0]=0.0;  };
						if(xt[1]>bound){xt[1]=bound;};
						if(xt[1]<0.0)  {xt[1]=0.0;  };
						q1 += xt[0];
						q2 += xt[1];
						w[j]=1.0;
						npe++;
					}else{
						w[j]=0.0;
					}
					x[j*2] = xt[0];
					x[j*2+1] = xt[1];
					j++;
				}
			}
			q1 /= npe;
			q2 /= npe;
			Nw = j;
		}
		if(verbose==1)printf("N of subjects available=%d  N of Mesh=%d\n", j, Nw);
		//for(j=0; j<N; j++){x[2*j]/=(q1+q2)/2.0; x[2*j+1]/=(q1+q2)/2.0;}
		if(Nw>1){
			settheta(Theta,q1,q2);
			GMM(x, w, Nw, Mu, Sigma, Theta, Theta15, Theta21, Z15, Z21, d, &lkhd, &status);
			
			for(j=0; j<2; j++){
				flag1=0;
				if(Theta[1]+Theta[4]<0.05 && Theta[1]+Theta[4]<Theta[3] && Theta[1]+Theta[4]<Theta[0]+Theta[2]+Theta[5]){
					cblas_dscal(K,Power(Mu[8]/Mu0[4],2),Sigma0,  4);
					cblas_dscal(K,Mu[8]/Mu0[4],         Sigma0+1,4);
					cblas_dscal(K,Mu[8]/Mu0[4],         Sigma0+2,4);
					cblas_dscal(K,Mu[8]/Mu0[4],Mu0,2);
					flag1 = 1;
					double tmp = Theta[3]; Theta[3]=Theta[1]; Theta[1]=tmp;
				}
				if(Theta[2]+Theta[4]<0.05 && Theta[2]+Theta[4]<Theta[5] && Theta[2]+Theta[4]<Theta[0]+Theta[1]+Theta[3]){
					cblas_dscal(K,Power(Mu[29]/Mu0[19],2),Sigma0+3,4);
					cblas_dscal(K,Mu[29]/Mu0[19],         Sigma0+1,4);
					cblas_dscal(K,Mu[29]/Mu0[19],         Sigma0+2,4);
					cblas_dscal(K,Mu[29]/Mu0[19],Mu0+1,2);
					flag1 = 1; 
					double tmp = Theta[5]; Theta[5]=Theta[2]; Theta[2]=tmp;
				}
				if(flag1==1){
					GMM(x, w, Nw, Mu, Sigma, Theta, Theta15, Theta21, Z15, Z21, d, &lkhd, &status);
				}
			}
			if(chkABtheta(Theta)==1){
				cblas_dscal(K,Power(Mu[19]/Mu0[19],2),Sigma0+3,4);
				cblas_dscal(K,Mu[19]/Mu0[19]*Mu[4]/Mu0[4],Sigma0+1,4);
				cblas_dscal(K,Mu[19]/Mu0[19]*Mu[4]/Mu0[4],Sigma0+2,4);
				cblas_dscal(K,Power(Mu[4]/Mu0[4],2),  Sigma0,  4);
				cblas_dscal(K,Mu[19]/Mu0[19],Mu0+1,2);
				cblas_dscal(K,Mu[4]/Mu0[4],Mu0,2);
				settheta(Theta,q1,q2);
				GMM(x, w, Nw, Mu, Sigma, Theta, Theta15, Theta21, Z15, Z21, d, &lkhd, &status);
			}
			cblas_dcopy(K*2,Mutemp,1,Mu0,1);
			cblas_dcopy(K*4,Sigmatemp,1,Sigma0,1);
			
			
			
			getETheta15(Theta,eTheta15);
			
if(verbose==1){
vprint(Mu,K*2);vprint(Sigma,K*4);vprint(Theta,J);//printf("l=%d ep=%d\n", chklattice(Mu,Sigma,eTheta15), chkEP(Mu,Sigma,eTheta15));printf("lkhd=%lf\n",lkhd);
}
			double pchisq=0.0;
			cr[0]=cr[1]=0.0;
			for(i=0; i<Nw; i++){ if(d[i]>0.0){cr[0]+=w[i];}; if(d[i]>(0.05/(double)N)){cr[1]+=w[i];} }
			getChisq(Theta15, Theta, cr[0], &pchisq);
			
			if(status==0){
				if(chklattice(Mu,Sigma,eTheta15)>0){
					if(chkEP(Mu,Sigma,eTheta15)>0){
						status = 3;
					}else{
						status = 1;
					}
				}else{
					if(chkEP(Mu,Sigma,eTheta15)>0){
						status = 2;
					}
				}
			}
			
			if(text==1){
				fprintf(op,"%d\t",itr+1);
				for(i=0; i<K*2; i++){fprintf(op,"%.5lf\t",Mu[i]);}
				for(i=0; i<K*4; i++){fprintf(op,"%.5lf\t",Sigma[i]);}
				for(i=0; i<K; i++){fprintf(op,"%.5lf\t",Theta15[i]);}
				for(i=0; i<J; i++){fprintf(op,"%.5lf\t",Theta[i]);}
				fprintf(op,"%d\t%.4lf\t%.2lf\t%.4lf\t%.2lf\t%lf\n",
							status,
							cr[1]/cr[0], 
							log10(pchisq), 
							getErrorRate(Z15, Theta15, w, Nw),
							lkhd, mesh_ratio);
			}else{
				double qcstats[]={
							(double)status, 
							cr[1]/cr[0], 
							log10(pchisq), 
							getErrorRate(Z15, Theta15, w, Nw),
							lkhd, mesh_ratio};
				fwrite(Mu, sizeof(double), K*2, op);
				fwrite(Sigma, sizeof(double), K*4, op);
				fwrite(Theta15, sizeof(double), K, op);
				fwrite(Theta, sizeof(double), J, op);
				fwrite(qcstats, sizeof(double), qcstatslen, op);
			}
			printf("\n");
		}else{printf("No data points!\n");
			if(text==1){
				fprintf(op,"%d\t",itr+1);
				for(i=0; i<K*7+J; i++){fprintf(op,"NA\t");}
				fprintf(op,"99\t");
				for(i=1; i<qcstatslen-1; i++){fprintf(op,"NA\t");}
				fprintf(op,"NA\n");
			}else{
				fwrite(Mu, sizeof(double), K*2, op);
				fwrite(Sigma, sizeof(double), K*4, op);
				fwrite(freq, sizeof(double), K, op);
				fwrite(Theta, sizeof(double), J, op);
				double qcstats[] = {99.,0.,0.,1.,0.,1.};
				fwrite(qcstats, sizeof(double), qcstatslen, op);
			}
		}
	}
	free(Z15);
	free(Z21);
	free(uvk);
	free(uvn);
	free(uvj);
	
	fclose(op);
	fclose(fp);
	
	return 0;
}














